#pylint: disable=missing-module-docstring, missing-function-docstring, line-too-long, missing-class-docstring
import os
from typing import Dict
from .utils import get_config_value, reread_files, disk_usage, get_media_file_list
from .file_convert import FileConvert

class DefaultStorageStrategy():
    """
    stores the latest {max_file_count} pictures in the {media_file} textfile
    all other pictures are written to the {complete_media_file} textfile
    """

    def __init__(self, config: Dict, logger) -> None:
        self.log = logger
        self.media_path = get_config_value(config, "media_path")
        self.media_file = get_config_value(config, "media_file")
        self.max_file_count = int(get_config_value(config, "max_file_count"))
        self.complete_media_file = get_config_value(config, "complete_media_file", False)
        self._convert = FileConvert(get_config_value(config["convert"], "convert_binary"), logger)
        self.convert_on_save = get_config_value(config["convert"], "convert_on_save")
        self.convert_parameters = get_config_value(config["convert"], "convert_parameters")
        self.min_free_disk_space_mb = get_config_value(config, "min_free_disk_space_mb")

    def _append_to_complete_media_file(self, filename) -> None:
        if not self.complete_media_file:
            return

        with open(self.complete_media_file, 'a', encoding='utf-8') as binary_file:
            binary_file.write(f'{filename}\n')

    def _add_to_media_file(self, filename) -> None:
        file_data = []
        try:
            with open(self.media_file, 'r', encoding='utf-8') as text_file:
                file_data = text_file.readlines()
        except IOError:
            pass

        file_data.append(f'{filename}\n')
        with open(self.media_file, 'w+', encoding='utf-8') as text_file:
            new_data = file_data[-self.max_file_count:]
            text_file.writelines(new_data)

    def _get_next_filename(self, prefered_filename: str, index: int=0) -> str:
        suffix = f" #{index}" if index > 0 else ''
        new_filename = f'{prefered_filename}{suffix}'
        if os.path.exists(new_filename):
            return self._get_next_filename(prefered_filename, index+1)
        return new_filename

    def _convert_file(self, filename: str):
        self._convert.convert_file(filename, self.convert_parameters)

    def _delete_eldest_file(self):
        try:
            file_list = get_media_file_list(self.media_path)
            if len(file_list) > 0:
                file_to_delete = file_list[0]
                os.remove(file_to_delete)
        #pylint: disable=broad-except
        except Exception as error:
            self.log.error(error)
        #pylint: enable=broad-except

    def _delete_eldest_files(self):
        free_space_mb = disk_usage(self.media_path).free / (1024 * 1024)
        if (self.min_free_disk_space_mb > 0) and self.min_free_disk_space_mb > free_space_mb:
            self._delete_eldest_file()
            self._delete_eldest_files()

    def _check_storage_limit(self):
        free_space_mb = disk_usage(self.media_path).free / (1024 * 1024)
        if (self.min_free_disk_space_mb > 0) and self.min_free_disk_space_mb > free_space_mb:
            self._delete_eldest_files()
            reread_files(self.media_path, self.media_file, self.complete_media_file, self.max_file_count)

    # Todo create images_local folder when not exists
    def store(self, data: bytes, filename: str) -> None:
        target = self._get_next_filename(os.path.join(self.media_path, filename))

        self._check_storage_limit()

        self.log.trace(f'save file as {target}')
        with open(target, "wb") as binary_file:
            binary_file.write(data)
            binary_file.close()

        if self.convert_on_save:
            self._convert_file(target)

        self._add_to_media_file(target)
        self._append_to_complete_media_file(target)
#pylint: enable=missing-module-docstring, missing-function-docstring, line-too-long, missing-class-docstring



#pylint: disable=missing-module-docstring, missing-function-docstring, line-too-long, missing-class-docstring
"""
    PhotOS Client

    This client uses the mautrix matrix framework available at:
    https://github.com/mautrix/python

    Parts from this source code are inspired by maubot:
    https://github.com/maubot/maubot
"""
import sys
import asyncio
import traceback
import random
from typing import Dict, Union
from mautrix.client import client as mau
from mautrix.crypto import PgCryptoStateStore, OlmMachine, StateStore as CryptoStateStore, PgCryptoStore
from mautrix.crypto.attachments.attachments import decrypt_attachment
from mautrix.client.state_store.sqlalchemy import SQLStateStore as BaseSQLStateStore
from mautrix.types.event.encrypted import EncryptedEvent
from mautrix.types.event.message import MediaMessageEventContent, MessageType, TextMessageEventContent
from mautrix.types.misc import PaginationDirection
from mautrix.types.primitive import RoomID, UserID
from mautrix.util.async_db import Database as AsyncDatabase

from mautrix.types import (StrippedStateEvent, Membership,
                           EventType)

from matrix_photos.text_message_command_handler import TextmessageCommandHandler

from .storage_strategy import DefaultStorageStrategy
from .utils import get_config_value
from .admin_command_handler import AdminCommandHandler
from .text_message_command_handler import TextmessageCommandHandler


class SQLStateStore(BaseSQLStateStore, CryptoStateStore):
    pass

class PhotOsClient():
    '''
        A Simple Matrix client which automatically joins room invitations from trusted users
        and downloads all attachments into a specified folder
    '''

    global_state_store: Union['BaseSQLStateStore', 'CryptoStateStore'] = SQLStateStore()

    def __init__(self, config: Dict, client_session, logger) -> None:

        self.user_id = get_config_value(config, "user_id")
        self.device_id = get_config_value(config, "device_id")
        self.base_url = get_config_value(config, "base_url")
        self.database_url = get_config_value(config, "database_url")
        self.user_password = get_config_value(config, "user_password")
        self.media_path = get_config_value(config, "media_path")
        self.trusted_users = get_config_value(config, "trusted_users")
        self.media_file = get_config_value(config, "media_file")
        self.client_session = client_session
        self.log = logger
        self.storage_strategy = DefaultStorageStrategy(config, logger)
        self.admin_user = get_config_value(config, "admin_user", False)
        self.allowed_mimetypes = get_config_value(config, "allowed_mimetypes", False)
        if self.admin_user:
            self.admin_command_handler = AdminCommandHandler(self.admin_user, config, logger)

        self.text_message_command_handler = TextmessageCommandHandler(config, logger)
        self.random_response_messages = get_config_value(config, "random_response_messages", False)

        self.crypto_db = None
        self.client = None

    async def initialize(self):
        '''Prepare crypto store and initialize a matrix client'''
        self.crypto_db = AsyncDatabase.create(self.database_url, upgrade_table=PgCryptoStore.upgrade_table)
        crypto_store = PgCryptoStore(account_id=self.user_id, pickle_key="mau.crypto", db=self.crypto_db)
        state_store = PgCryptoStateStore(self.crypto_db)

        self.client = mau.Client(mxid=self.user_id,
                base_url=self.base_url,
                device_id=self.device_id,
                client_session=self.client_session,
                state_store=state_store,
                sync_store=crypto_store,
                log=self.log)

        await self.crypto_db.start()
        await state_store.upgrade_table.upgrade(self.crypto_db)
        await crypto_store.open()

        crypto = OlmMachine(self.client, crypto_store, state_store, self.log)

        self.client.crypto = crypto
        self.client.crypto_log = self.log

        self.client.ignore_first_sync = False
        self.client.ignore_initial_sync = False

        crypto_device_id = await crypto_store.get_device_id()
        if crypto_device_id and crypto_device_id != self.device_id:
            self.log.fatal("Mismatching device ID in crypto store and config "
                      f"(store: {crypto_device_id}, config: {self.device_id})")
            sys.exit(10)

        await self.client.crypto.load()
        if not crypto_device_id:
            await crypto_store.put_device_id(self.device_id)

        if self.client.crypto_enabled:
            self.log.debug("Enabled encryption support")

        login_response = await self.client.login(self.user_id, password=self.user_password)
        self.log.trace(login_response)

        while True:
            try:
                whoami = await self.client.whoami()
            #pylint: disable=broad-except
            except Exception:
                self.log.exception("Failed to connect to homeserver, retrying in 10 seconds...")
                await asyncio.sleep(10)
                continue
            #pylint: enable=broad-except
            if whoami.user_id != self.user_id:
                self.log.fatal(f"User ID mismatch: configured {self.user_id}, but server said {whoami.user_id}")
                sys.exit(11)
            elif whoami.device_id and self.device_id and whoami.device_id != self.device_id:
                self.log.fatal(f"Device ID mismatch: configured {self.device_id}, "
                        f"but server said {whoami.device_id}")
                sys.exit(12)
            self.log.debug(f"Confirmed connection as {whoami.user_id} / {whoami.device_id}")
            break

        self.client.add_event_handler(EventType.ROOM_MEMBER, self._handle_invite)
        self.client.add_event_handler(EventType.ROOM_MESSAGE, self._handle_message)

    async def _handle_invite(self, evt: StrippedStateEvent) -> None:
        self.log.trace('_handle_invite')
        self.log.trace(evt.state_key)

        if evt.state_key == self.user_id and self.is_trusted_user(evt.sender) and evt.content.membership == Membership.INVITE:
            self.log.debug('join room!')
            await self.client.join_room(evt.room_id)

        if evt.state_key == self.user_id and not self.is_trusted_user(evt.sender) and evt.content.membership == Membership.INVITE:
            self.log.trace(f'untrusted user {evt.sender}')

    def is_trusted_user(self, user_id: UserID) -> bool:
        if not user_id:
            return False

        return user_id in self.trusted_users

    def is_admin_user(self, user_id: UserID) -> bool:
        return user_id == self.admin_user

    async def _store_data(self, media_content: MediaMessageEventContent) -> None:
        encrypted_data = await self.client.download_media(media_content.file.url)

        file_hash = media_content.file.hashes['sha256']
        vector = media_content.file.iv
        decrypted_data = decrypt_attachment(encrypted_data, media_content.file.key.key, file_hash, vector)

        #IDEA maybe store the hash somewhere and only store the file if we dont have a file with the same hash
        self.storage_strategy.store(decrypted_data, str(media_content.body))


    def _is_allowed_content(self, content: MediaMessageEventContent):
        result = content.info.mimetype in self.allowed_mimetypes
        if not result:
            self.log.warn(f'mimetype not allowed: {content.info.mimetype}')
        return result

    async def _handle_admin_command(self, evt: StrippedStateEvent):
        if self.admin_command_handler:
            response = self.admin_command_handler.handle(evt.content)
            if response:
                content = TextMessageEventContent(MessageType.TEXT, response)
                content.set_reply(evt)
                await self.client.send_message_event(evt.room_id, EventType.ROOM_MESSAGE, content)

    def _is_admin_command(self, evt: StrippedStateEvent) -> bool:
        if self.is_admin_user(evt.sender) and self.admin_command_handler:
            return self.admin_command_handler.is_admin_command(evt.content)

        return False

    async def _handle_message_event(self, evt: StrippedStateEvent) -> None:
        if self._is_admin_command(evt):
            await self._handle_admin_command(evt)
        else:
            is_foreign_message = evt.sender != self.user_id
            media_message_before = await self.message_before_was_media_message(evt.room_id)
            if is_foreign_message and media_message_before:
                self.text_message_command_handler.handle(evt.content)


    async def message_before_was_media_message(self, room_id: RoomID) -> bool:
        token = await self.client.sync_store.get_next_batch()
        if token:
            messages = await self.client.get_messages(room_id, direction=PaginationDirection.BACKWARD, from_token=token, limit=10)
            foreign_messages = list(filter(lambda x: x.sender != self.user_id, messages.events))[:2]

            if len(foreign_messages) == 2:
                event = foreign_messages[1]
                if isinstance(event, EncryptedEvent):
                    decrypted = await self.client.crypto.decrypt_megolm_event(event)
                    if decrypted and isinstance(decrypted.content, MediaMessageEventContent):
                        return True
        return False

    async def _send_random_response_message(self, evt: StrippedStateEvent):
        if not self.random_response_messages:
            return

        response = random.choice(self.random_response_messages)
        content = TextMessageEventContent(MessageType.TEXT, response)
        content.set_reply(evt)
        await self.client.send_message_event(evt.room_id, EventType.ROOM_MESSAGE, content)

    async def _handle_message(self, evt: StrippedStateEvent) -> None:
        self.log.trace('_handle_message')

        try:
            if isinstance(evt.content, TextMessageEventContent):
                self.log.trace('TextMessageEventContent')
                await self._handle_message_event(evt)

            if isinstance(evt.content, MediaMessageEventContent) and self._is_allowed_content(evt.content):
                self.log.trace('MediaMessageEventContent')
                await self._store_data(evt.content)
                await self._send_random_response_message(evt)

        #pylint: disable=broad-except
        except Exception as error:
            self.log.error(error)
            traceback.print_exc()
        #pylint: enable=broad-except

    async def stop(self):
        self.client.stop()
        await self.crypto_db.stop()
        self.log.info('client stopped!')

    async def start(self):
        self.log.info('starting client')
        self.client.start(None)
#pylint: enable=missing-module-docstring, missing-function-docstring, line-too-long, missing-class-docstring



#pylint: disable=missing-module-docstring, missing-function-docstring, line-too-long, missing-class-docstring, broad-except
import subprocess

class FileConvert:

    def __init__(self, convert_binary: str, logger) -> None:
        self.log = logger
        self.convert_binary = convert_binary

    def convert_file(self, filename, convert_params, message=None, convert_text_parameter: str=None):
        try:
            self.log.trace(f'convert_file {filename}')

            text = []

            if message:
                msg = f"'{message}'"
                text.append(f'{convert_text_parameter} {msg}')

            params = [
                self.convert_binary,
                *convert_params,
                *text,
                f'{filename}',
                f'{filename}'
                ]

            result = subprocess.run(params, capture_output=True, text=True, check=True)
            self.log.trace(result.stdout)
            self.log.trace(result.stderr)
        except Exception as error:
            self.log.error(error)
#pylint: enable=missing-module-docstring, missing-function-docstring, line-too-long, missing-class-docstring, broad-except



'''
Photos Matrix File Download
'''
from typing import cast
import logging.config
import asyncio
import signal
import sys
import argparse
import yaml
from aiohttp import ClientSession
from mautrix.util.logging import TraceLogger
from .photos_client import PhotOsClient

#pylint: disable=global-statement, missing-function-docstring, broad-except, global-variable-not-assigned
loop = asyncio.get_event_loop()

commandline_parser = argparse.ArgumentParser(description="PhotOS Matrix client")
commandline_parser.add_argument("-c", "--config", type=str,
                                default="~/config.yaml",
                                required=True, metavar="<path>",
                                help="the path to your config file")

args = commandline_parser.parse_args()

PHOTOS_CLIENT = None
HTTP_CLIENT = None
CONFIG = None

with open(args.config, 'r', encoding='utf-8') as stream:
    CONFIG = yaml.load(stream, Loader=yaml.FullLoader)

logging.config.dictConfig(CONFIG["logging"])
logger = logging.getLogger(__name__)
TRACE_LOGGER = cast(TraceLogger, logger)

async def main():
    global HTTP_CLIENT
    HTTP_CLIENT = ClientSession(loop=loop)
    global PHOTOS_CLIENT

    async def try_connect() -> bool:
        try:
            global PHOTOS_CLIENT
            global TRACE_LOGGER
            global CONFIG
            await PHOTOS_CLIENT.initialize()
            await PHOTOS_CLIENT.start()
            return True
        except Exception as exception:
            TRACE_LOGGER.exception(exception)
            return False

    PHOTOS_CLIENT = PhotOsClient(CONFIG["matrix"], HTTP_CLIENT, TRACE_LOGGER)

    while not await try_connect():
        print('failure')
        await asyncio.sleep(5)

async def stop() -> None:
    logger.info('terminate client')
    if PHOTOS_CLIENT:
        await PHOTOS_CLIENT.stop()

    if HTTP_CLIENT:
        await HTTP_CLIENT.close()

try:
    logger.info("Starting PhotOS Matrix Client")
    loop.run_until_complete(main())
except Exception:
    logger.fatal("Failed to start plugin", exc_info=True)
    loop.run_until_complete(stop())
    loop.close()
    sys.exit(1)

signal.signal(signal.SIGINT, signal.default_int_handler)
signal.signal(signal.SIGTERM, signal.default_int_handler)

try:
    logger.info("Startup completed, running forever")
    loop.run_forever()
except KeyboardInterrupt:
    logger.info("Interrupt received, stopping")
    loop.run_until_complete(stop())
    loop.close()
    sys.exit(0)

except Exception:
    logger.fatal("Fatal error in PhotOS Matrix Client", exc_info=True)
    sys.exit(1)
#pylint: enable=global-statement, missing-function-docstring, broad-except, global-variable-not-assigned



#pylint: disable=missing-module-docstring, missing-function-docstring, line-too-long, missing-class-docstring
from enum import Enum
from typing import Dict, List, Tuple
from mautrix.types.event.message import MessageType, TextMessageEventContent
from .utils import get_config_value, disk_usage, reread_files

class AdminCommands(str, Enum):
    HELP = '!help'
    REREAD = '!reread'
    STATS = '!stats'

    @staticmethod
    def list():
        return list(map(lambda c: c.value, AdminCommands))

    @staticmethod
    def get_description(command):
        if command == AdminCommands.HELP:
            return f'{command} - shows this message'
        if command == AdminCommands.REREAD:
            return f'{command} - reread directory with images and create image text files'
        if command == AdminCommands.STATS:
            return f'{command} - show various statistics like free diskspace'

    @staticmethod
    def help_message():
        return list(map(AdminCommands.get_description, AdminCommands))

class AdminCommandHandler:

    def __init__(self, admin_user: str, config: Dict, logger) -> None:
        self.log = logger
        self.admin_user = admin_user
        self.media_path = get_config_value(config, "media_path")
        self.media_file = get_config_value(config, "media_file")
        self.max_file_count = int(get_config_value(config, "max_file_count"))
        self.complete_media_file = get_config_value(config, "complete_media_file", False)

    def _create_help_message(self) -> str:
        return "\n".join(AdminCommands.help_message())

    def _show_stats(self) -> str:
        stats = disk_usage(self.media_path)
        free_mb = stats.free / (1024*1024*1024)
        return f'Free disk space (Gb): {free_mb}'

    def _reread_files(self) -> str:
        reread_files(self.media_path, self.media_file, self.complete_media_file, self.max_file_count)
        return "Done reread files"

    def _handle_command(self, command: str, params: List) -> str:
        self.log.trace(f'_handle_command: {command}')
        self.log.trace(params)
        try:
            if command == AdminCommands.REREAD:
                return self._reread_files()
            if command == AdminCommands.HELP:
                return self._create_help_message()
            if command == AdminCommands.STATS:
                return self._show_stats()
        #pylint: disable=broad-except
        except Exception as exception:
            return str(exception)
        #pylint: enable=broad-except

        return None

    def _get_command_with_parameters(self, content: TextMessageEventContent) -> Tuple[str, List[str]]:
        commands = content.body.split(' ')
        if len(commands) <= 0:
            self.log.trace("no commands given")
            return None

        return (commands[0], commands[1:] if len(commands) > 1 else [])

    def is_admin_command(self, content: TextMessageEventContent):
        if content.msgtype == MessageType.TEXT:
            commands = self._get_command_with_parameters(content)
            if commands:
                return commands[0].startswith("!")

        return False

    def handle(self, content: TextMessageEventContent):
        self.log.trace('handle admincommand')
        if self.is_admin_command(content):
            commands = self._get_command_with_parameters(content)
            return self._handle_command(commands[0], commands[1])

        return None
#pylint: enable=missing-module-docstring, missing-function-docstring, line-too-long, missing-class-docstring



#pylint: disable=missing-module-docstring, missing-function-docstring, line-too-long, missing-class-docstring, broad-except
import subprocess
from typing import Dict
from mautrix.types.event.message import MessageType, TextMessageEventContent
from .utils import get_config_value
from .file_convert import FileConvert
class TextmessageCommandHandler:

    def __init__(self, config: Dict, logger) -> None:
        self.log = logger
        self.write_text_messages = get_config_value(config["message_convert"], "write_text_messages")
        self.convert_parameters = get_config_value(config["message_convert"], "convert_parameters")
        self.convert_text_parameter = get_config_value(config["message_convert"], "convert_text_parameter")
        self.media_file = get_config_value(config, "media_file")
        self._convert = FileConvert(get_config_value(config["message_convert"], "convert_binary"), logger)

    def _get_last_filename(self):
        file_data = []
        try:
            with open(self.media_file, 'r', encoding='utf-8') as text_file:
                file_data = text_file.readlines()
                return file_data[-1]
        except IOError:
            pass
        return ""

    # ToDo better exception handling when command fails
    def _add_message_to_file(self, filename, message):
        self.log.trace(f'_add_text_to_file {message}')
        self._convert.convert_file(filename, self.convert_parameters, message=message, convert_text_parameter=self.convert_text_parameter)

    def _handle_text_message(self, content: TextMessageEventContent):
        target_filename = str(self._get_last_filename()).strip()
        self.log.trace('_handle_text_message')
        if target_filename:
            self._add_message_to_file(target_filename, content.body)

    def handle(self, content: TextMessageEventContent):
        try:
            if content.msgtype == MessageType.TEXT and not content.body.startswith('!') and self.write_text_messages:
                return self._handle_text_message(content)

            return None
        except Exception as error:
            self.log.error(error)
            return None
#pylint: enable=missing-module-docstring, missing-function-docstring, line-too-long, missing-class-docstring, broad-except
